@@ -12,10 +12,13 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from collections import defaultdict
+from __future__ import absolute_import  # to enable import io from stdlib
+from collections import defaultdict, deque
 import errno
 from functools import wraps, partial
+import io
 import logging
+import os
 import sys
 from threading import Event, RLock
 import time
@@ -29,13 +32,13 @@ import six
 from six.moves import range
 
 from cassandra import ConsistencyLevel, AuthenticationFailed, OperationTimedOut
-from cassandra.marshal import int32_pack, header_unpack
+from cassandra.marshal import int32_pack, header_unpack, v3_header_unpack, int32_unpack
 from cassandra.protocol import (ReadyMessage, AuthenticateMessage, OptionsMessage,
                                 StartupMessage, ErrorMessage, CredentialsMessage,
                                 QueryMessage, ResultMessage, decode_response,
                                 InvalidRequestException, SupportedMessage,
                                 AuthResponseMessage, AuthChallengeMessage,
-                                AuthSuccessMessage)
+                                AuthSuccessMessage, ProtocolException)
 from cassandra.util import OrderedDict
 
 
@@ -79,8 +82,6 @@ else:
     locally_supported_compressions['snappy'] = (snappy.compress, decompress)
 
 
-MAX_STREAM_PER_CONNECTION = 127
-
 PROTOCOL_VERSION_MASK = 0x7f
 
 HEADER_DIRECTION_FROM_CLIENT = 0x00
@@ -152,16 +153,32 @@ class Connection(object):
 
     ssl_options = None
     last_error = None
+
+    # The current number of operations that are in flight. More precisely,
+    # the number of request IDs that are currently in use.
     in_flight = 0
+
+    # A set of available request IDs.  When using the v3 protocol or higher,
+    # this will no initially include all request IDs in order to save memory,
+    # but the set will grow if it is exhausted.
+    request_ids = None
+
+    # Tracks the highest used request ID in order to help with growing the
+    # request_ids set
+    highest_request_id = 0
+
     is_defunct = False
     is_closed = False
     lock = None
+    user_type_map = None
 
     is_control_connection = False
+    _iobuf = None
 
     def __init__(self, host='127.0.0.1', port=9042, authenticator=None,
                  ssl_options=None, sockopts=None, compression=True,
-                 cql_version=None, protocol_version=2, is_control_connection=False):
+                 cql_version=None, protocol_version=2, is_control_connection=False,
+                 user_type_map=None):
         self.host = host
         self.port = port
         self.authenticator = authenticator
@@ -171,11 +188,36 @@ class Connection(object):
         self.cql_version = cql_version
         self.protocol_version = protocol_version
         self.is_control_connection = is_control_connection
+        self.user_type_map = user_type_map
         self._push_watchers = defaultdict(set)
+        self._iobuf = io.BytesIO()
+        if protocol_version >= 3:
+            self._header_unpack = v3_header_unpack
+            self._header_length = 5
+            self.max_request_id = (2 ** 15) - 1
+            # Don't fill the deque with 2**15 items right away. Start with 300 and add
+            # more if needed.
+            self.request_ids = deque(range(300))
+            self.highest_request_id = 299
+        else:
+            self._header_unpack = header_unpack
+            self._header_length = 4
+            self.max_request_id = (2 ** 7) - 1
+            self.request_ids = deque(range(self.max_request_id + 1))
+            self.highest_request_id = self.max_request_id
 
-        self._id_queue = Queue(MAX_STREAM_PER_CONNECTION)
-        for i in range(MAX_STREAM_PER_CONNECTION):
-            self._id_queue.put_nowait(i)
+        # 0         8        16        24        32         40
+        # +---------+---------+---------+---------+---------+
+        # | version |  flags  |      stream       | opcode  |
+        # +---------+---------+---------+---------+---------+
+        # |                length                 |
+        # +---------+---------+---------+---------+
+        # |                                       |
+        # .            ...  body ...              .
+        # .                                       .
+        # .                                       .
+        # +----------------------------------------
+        self._full_header_length = self._header_length + 4
 
         self.lock = RLock()
 
@@ -187,6 +229,14 @@ class Connection(object):
         """
         pass
 
+    @classmethod
+    def handle_fork(self):
+        """
+        Called after a forking.  This should cleanup any remaining reactor state
+        from the parent process.
+        """
+        pass
+
     def close(self):
         raise NotImplementedError()
 
@@ -218,6 +268,18 @@ class Connection(object):
                             "failed connection (%s) to host %s:",
                             id(self), self.host, exc_info=True)
 
+    def get_request_id(self):
+        """
+        This must be called while self.lock is held.
+        """
+        try:
+            return self.request_ids.popleft()
+        except IndexError:
+            self.highest_request_id += 1
+            # in_flight checks should guarantee this
+            assert self.highest_request_id <= self.max_request_id
+            return self.highest_request_id
+
     def handle_pushed(self, response):
         log.debug("Message pushed from server: %r", response)
         for cb in self._push_watchers.get(response.event_type, []):
@@ -226,21 +288,12 @@ class Connection(object):
             except Exception:
                 log.exception("Pushed event handler errored, ignoring:")
 
-    def send_msg(self, msg, cb, wait_for_id=False):
+    def send_msg(self, msg, request_id, cb):
         if self.is_defunct:
             raise ConnectionShutdown("Connection to %s is defunct" % self.host)
         elif self.is_closed:
             raise ConnectionShutdown("Connection to %s is closed" % self.host)
 
-        if not wait_for_id:
-            try:
-                request_id = self._id_queue.get_nowait()
-            except Empty:
-                raise ConnectionBusy(
-                    "Connection to %s is at the max number of requests" % self.host)
-        else:
-            request_id = self._id_queue.get()
-
         self._callbacks[request_id] = cb
         self.push(msg.to_binary(request_id, self.protocol_version, compression=self.compressor))
         return request_id
@@ -249,21 +302,33 @@ class Connection(object):
         return self.wait_for_responses(msg, timeout=timeout)[0]
 
     def wait_for_responses(self, *msgs, **kwargs):
+        """
+        Returns a list of (success, response) tuples.  If success
+        is False, response will be an Exception.  Otherwise, response
+        will be the normal query response.
+
+        If fail_on_error was left as True and one of the requests
+        failed, the corresponding Exception will be raised.
+        """
         if self.is_closed or self.is_defunct:
             raise ConnectionShutdown("Connection %s is already closed" % (self, ))
         timeout = kwargs.get('timeout')
-        waiter = ResponseWaiter(self, len(msgs))
+        fail_on_error = kwargs.get('fail_on_error', True)
+        waiter = ResponseWaiter(self, len(msgs), fail_on_error)
 
         # busy wait for sufficient space on the connection
         messages_sent = 0
         while True:
             needed = len(msgs) - messages_sent
             with self.lock:
-                available = min(needed, MAX_STREAM_PER_CONNECTION - self.in_flight)
+                available = min(needed, self.max_request_id - self.in_flight)
+                request_ids = [self.get_request_id() for _ in range(available)]
                 self.in_flight += available
 
-            for i in range(messages_sent, messages_sent + available):
-                self.send_msg(msgs[i], partial(waiter.got_response, index=i), wait_for_id=True)
+            for i, request_id in enumerate(request_ids):
+                self.send_msg(msgs[messages_sent + i],
+                              request_id,
+                              partial(waiter.got_response, index=messages_sent + i))
             messages_sent += available
 
             if messages_sent == len(msgs):
@@ -293,14 +358,48 @@ class Connection(object):
         self.is_control_connection = False
         self._push_watchers = {}
 
+    def process_io_buffer(self):
+        while True:
+            pos = self._iobuf.tell()
+            if pos < self._full_header_length or (self._total_reqd_bytes > 0 and pos < self._total_reqd_bytes):
+                # we don't have a complete header yet or we
+                # already saw a header, but we don't have a
+                # complete message yet
+                return
+            else:
+                # have enough for header, read body len from header
+                self._iobuf.seek(self._header_length)
+                body_len = int32_unpack(self._iobuf.read(4))
+
+                # seek to end to get length of current buffer
+                self._iobuf.seek(0, os.SEEK_END)
+                pos = self._iobuf.tell()
+
+                if pos >= body_len + self._full_header_length:
+                    # read message header and body
+                    self._iobuf.seek(0)
+                    msg = self._iobuf.read(self._full_header_length + body_len)
+
+                    # leave leftover in current buffer
+                    leftover = self._iobuf.read()
+                    self._iobuf = io.BytesIO()
+                    self._iobuf.write(leftover)
+
+                    self._total_reqd_bytes = 0
+                    self.process_msg(msg, body_len)
+                else:
+                    self._total_reqd_bytes = body_len + self._full_header_length
+                    return
+
     @defunct_on_error
     def process_msg(self, msg, body_len):
-        version, flags, stream_id, opcode = header_unpack(msg[:4])
+        version, flags, stream_id, opcode = self._header_unpack(msg[:self._header_length])
         if stream_id < 0:
             callback = None
         else:
             callback = self._callbacks.pop(stream_id, None)
-            self._id_queue.put_nowait(stream_id)
+            with self.lock:
+                self.request_ids.append(stream_id)
 
         body = None
         try:
@@ -318,13 +417,14 @@ class Connection(object):
                     % (opcode, stream_id))
 
             if body_len > 0:
-                body = msg[8:]
+                body = msg[self._full_header_length:]
             elif body_len == 0:
                 body = six.binary_type()
             else:
                 raise ProtocolError("Got negative body length: %r" % body_len)
 
-            response = decode_response(stream_id, flags, opcode, body, self.decompressor)
+            response = decode_response(given_version, self.user_type_map, stream_id,
+                                       flags, opcode, body, self.decompressor)
         except Exception as exc:
             log.exception("Error decoding response from Cassandra. "
                           "opcode: %04x; message contents: %r", opcode, msg)
@@ -334,10 +434,14 @@ class Connection(object):
             return
 
         try:
-            if stream_id < 0:
+            if stream_id >= 0:
+                if isinstance(response, ProtocolException):
+                    log.error("Closing connection %s due to protocol error: %s", self, response.summary_msg())
+                    self.defunct(response)
+                if callback is not None:
+                    callback(response)
+            else:
                 self.handle_pushed(response)
-            elif callback is not None:
-                callback(response)
         except Exception:
             log.exception("Callback handler errored, ignoring:")
 
@@ -352,7 +456,7 @@ class Connection(object):
             self._send_startup_message()
         else:
             log.debug("Sending initial options message for new connection (%s) to %s", id(self), self.host)
-            self.send_msg(OptionsMessage(), self._handle_options_response)
+            self.send_msg(OptionsMessage(), self.get_request_id(), self._handle_options_response)
 
     @defunct_on_error
     def _handle_options_response(self, options_response):
@@ -363,10 +467,10 @@ class Connection(object):
             if isinstance(options_response, ConnectionException):
                 raise options_response
             else:
-                log.error("Did not get expected SupportedMessage response; " \
+                log.error("Did not get expected SupportedMessage response; "
                           "instead, got: %s", options_response)
-                raise ConnectionException("Did not get expected SupportedMessage " \
-                                          "response; instead, got: %s" \
+                raise ConnectionException("Did not get expected SupportedMessage "
+                                          "response; instead, got: %s"
                                           % (options_response,))
 
         log.debug("Received options response on new connection (%s) from %s",
@@ -419,11 +523,13 @@ class Connection(object):
 
     @defunct_on_error
     def _send_startup_message(self, compression=None):
+        log.debug("Sending StartupMessage on %s", self)
         opts = {}
         if compression:
             opts['COMPRESSION'] = compression
         sm = StartupMessage(cqlversion=self.cql_version, options=opts)
-        self.send_msg(sm, cb=self._handle_startup_response)
+        self.send_msg(sm, self.get_request_id(), cb=self._handle_startup_response)
+        log.debug("Sent StartupMessage on %s", self)
 
     @defunct_on_error
     def _handle_startup_response(self, startup_response, did_authenticate=False):
@@ -447,12 +553,12 @@ class Connection(object):
                 log.debug("Sending credentials-based auth response on %s", self)
                 cm = CredentialsMessage(creds=self.authenticator)
                 callback = partial(self._handle_startup_response, did_authenticate=True)
-                self.send_msg(cm, cb=callback)
+                self.send_msg(cm, self.get_request_id(), cb=callback)
             else:
                 log.debug("Sending SASL-based auth response on %s", self)
                 initial_response = self.authenticator.initial_response()
                 initial_response = "" if initial_response is None else initial_response.encode('utf-8')
-                self.send_msg(AuthResponseMessage(initial_response), self._handle_auth_response)
+                self.send_msg(AuthResponseMessage(initial_response), 0, self._handle_auth_response)
         elif isinstance(startup_response, ErrorMessage):
             log.debug("Received ErrorMessage on new connection (%s) from %s: %s",
                       id(self), self.host, startup_response.summary_msg())
@@ -487,7 +593,7 @@ class Connection(object):
             response = self.authenticator.evaluate_challenge(auth_response.challenge)
             msg = AuthResponseMessage("" if response is None else response)
             log.debug("Responding to auth challenge on %s", self)
-            self.send_msg(msg, self._handle_auth_response)
+            self.send_msg(msg, self.get_request_id(), self._handle_auth_response)
         elif isinstance(auth_response, ErrorMessage):
             log.debug("Received ErrorMessage on new connection (%s) from %s: %s",
                       id(self), self.host, auth_response.summary_msg())
@@ -551,7 +657,21 @@ class Connection(object):
                 callback(self, self.defunct(ConnectionException(
                     "Problem while setting keyspace: %r" % (result,), self.host)))
 
-        self.send_msg(query, process_result, wait_for_id=True)
+        request_id = None
+        # we use a busy wait on the lock here because:
+        # - we'll only spin if the connection is at max capacity, which is very
+        #   unlikely for a set_keyspace call
+        # - it allows us to avoid signaling a condition every time a request completes
+        while True:
+            with self.lock:
+                if self.in_flight < self.max_request_id:
+                    request_id = self.get_request_id()
+                    self.in_flight += 1
+                    break
+
+            time.sleep(0.001)
+
+        self.send_msg(query, request_id, process_result)
 
     def __str__(self):
         status = ""
@@ -566,9 +686,10 @@ class Connection(object):
 
 class ResponseWaiter(object):
 
-    def __init__(self, connection, num_responses):
+    def __init__(self, connection, num_responses, fail_on_error):
         self.connection = connection
         self.pending = num_responses
+        self.fail_on_error = fail_on_error
         self.error = None
         self.responses = [None] * num_responses
         self.event = Event()
@@ -577,15 +698,33 @@ class ResponseWaiter(object):
         with self.connection.lock:
             self.connection.in_flight -= 1
         if isinstance(response, Exception):
-            self.error = response
-            self.event.set()
-        else:
-            self.responses[index] = response
-            self.pending -= 1
-            if not self.pending:
+            if hasattr(response, 'to_exception'):
+                response = response.to_exception()
+            if self.fail_on_error:
+                self.error = response
                 self.event.set()
+            else:
+                self.responses[index] = (False, response)
+        else:
+            if not self.fail_on_error:
+                self.responses[index] = (True, response)
+            else:
+                self.responses[index] = response
+
+        self.pending -= 1
+        if not self.pending:
+            self.event.set()
 
     def deliver(self, timeout=None):
+        """
+        If fail_on_error was set to False, a list of (success, response)
+        tuples will be returned.  If success is False, response will be
+        an Exception.  Otherwise, response will be the normal query response.
+
+        If fail_on_error was left as True and one of the requests
+        failed, the corresponding Exception will be raised. Otherwise,
+        the normal response will be returned.
+        """
         self.event.wait(timeout)
         if self.error:
             raise self.error
